// Tom's Spectrum Analyzer - (c) 2001 by Tom Grandgent
// tom@grandgent.com
// Written on 1/8/2001
//
// Note 1: This code is kind of a mess!  Please feel free to ask questions.
//
// Note 2: This program uses FFT code by Dominic Mazzoni (see FFT-fast.cpp)

#include <windows.h>
#include <stdio.h>
#include <process.h>
#include <math.h>
#include <dsound.h>

#include "resource.h"
#include "FFT-fast.h"

#define SAFE_RELEASE(p) { if(p) { (p)->Release(); (p)=NULL; } }

#define FREQ_DIVISOR		(65536 * 4)

// System menu ID used for always-on-top menu item
#define SC_ON_TOP			1


// DECLARATIONS

// Windows variables
HWND hwndApp = NULL;			// Application window handle
HINSTANCE ghInstance = NULL;	// Application instance handle
HANDLE quitReqEvent = NULL;		// Event signalled when the program wants to die
HANDLE quitAckEvent = NULL;		// Event signalled when the analysis thread is finished
int pause_display = 0;			// Should the display be paused?
int AlwaysOnTop = 0;			// Should the window always be on top?

// DirectX variables
LPDIRECTSOUNDCAPTURE8 m_dsCap = NULL;
DSCCAPS m_dsCapCaps;
DSCBUFFERDESC m_dsCapBufDesc;
LPDIRECTSOUNDCAPTUREBUFFER m_dsCapBuf = NULL;
//WAVEFORMATEX m_dsWaveFormat = { WAVE_FORMAT_PCM, 1, 8000, 8000, 1, 8, 0 };
WAVEFORMATEX m_dsWaveFormat = { WAVE_FORMAT_PCM, 2, 44100, 176400, 4, 16, 0 };
//WAVEFORMATEX m_dsWaveFormat = { WAVE_FORMAT_PCM, 2, 22050, 88200, 4, 16, 0 };
// wFormatTag, nChannels, nSamplesPerSec, mAvgBytesPerSec, nBlockAlign, wBitsPerSample, cbSize
DSCBCAPS m_dsCapBufCaps;

// Application variables
char *baseTitle = "Tom's Spectrum Analyzer";
int drawWaves = 1;				// Draw the raw waveforms?
int drawVoicePrint = 0;			// Draw voiceprint?
float voicePrintZoom = 1.0f;	// Voiceprint zoom level
unsigned int capBufLen;			// Capture buffer length in bytes
DWORD RequestedLockLen = 4096;	// Bytes to lock in the sample buffer
DWORD lockSamples = 0;			// Samples locked (containing both left and right channels)
DWORD channelSamples = 0;		// Samples locked per channel
DWORD RequestedFFTSamples = RequestedLockLen / 4;	// Samples run through the FFT (zero-padded at end if necessary)
DWORD filterSamples = 0;		// Samples to use for filtering
float freqMultiplier = 1.0f;	// Frequency sample multiplier
int logleft = 0, logright = 0;	// Use a logarithmic display for left & right channels?
DWORD lastFrameTime = 0;

int selectionInProgress = 0;
int selectionMade = 0;
float selectionStart;
float selectionEnd;

// Data buffers for sample data - allocated/reallocated on demand
float *floatLeft = NULL;
float *floatRight = NULL;
float *floatImag = NULL;
float *floatLeftOut = NULL;
float *floatRightOut = NULL;
float *floatImagOut = NULL;
float *floatLeftMag = NULL;
float *floatRightMag = NULL;
float *floatLeftMax = NULL;
float *floatRightMax = NULL;

char *helpText = "This program uses DirectSoundCapture to capture an audio signal and display it in two ways:\n\n"
				 "Left: Spectrum Analysis (frequency domain)   Right: Raw Waveform (time domain)\n\n"
				 "The display is in stereo, with the left channel above the right channel.\n\n"
				 "You can control various parameters with the following keys: (use numeric keypad when possible)\n\n"
				 "[, ]\t\tLower/Raise Sample Buffer Length:  # of samples acquired per frame per channel.\n"
				 "/, *\t\tLower/Raise FFT Length:  # of samples run through the FFT. (zero-padded if needed)\n"
				 "PgDn,PgUp\tLower/Raise Filter Samples:  # of samples averaged to filter the FFT output.\n"
				 "-, +\t\tLower/Raise Frequency Multiplier:  Adjusts amplification of the spectrum analysis.\n"
				 "W\t\tToggle Waveform Display:  Gives more room for spectrum analysis.\n"
				 ";, '\t\tToggle logarithmic (decibel) display for left/right channels.\n"
				 "V\t\tToggle Voiceprint Display:  Shows a scrolling graph of frequency data.\n"
				 "A, Z\t\tLower/Raise Voiceprint Zoom:  Zooms in and out of the lower frequencies.\n"
				 "LMouseButton\tHold down over spectrum analysis to see specific frequency value in title bar.\n"
				 "RMouseButton\tClick to freeze display and show both filtered/unfiltered spectrum analysis.\n"
				 "\t\tNote that you can still adjust parameters while the display is frozen.\n\n"
				 "This program was written by Tom Grandgent on Jan. 8th, 2001 and refined over the next few nights.\n"
				 "The objective was to try something with digital signal processing after reading most of the\n"
				 "excellent book, \"The Scientist and Engineer's Guide to Digital Signal Processing\".  Thanks Dad!\n\n"
				 "I can be reached online at: tom@grandgent.com and http://www.grandgent.com/tom/\n\n"
				 "Note: This program uses optimized FFT code by Dominic Mazzoni.";

// Application functions
int init_directsound_capture(void);
void shutdown_directsound_capture(void);

// Threads
void _cdecl analyze_thread(void *ptr);

// Dialog handlers
BOOL CALLBACK MainDlg(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam);


// IMPLEMENTATION

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, 
				   PSTR szCmdLine, int iCmdShow)
{
	ghInstance = hInstance;

	if (!init_directsound_capture())
		DialogBox(ghInstance, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC) MainDlg);
	else
		MessageBox(NULL, "Initialization failed.  This program requires DirectX 8 and a sound card capable of capturing audio at 44KHz 16-bit stereo using DirectSoundCapture.", "TSpec says.. Sorry!", MB_ICONERROR);

	shutdown_directsound_capture();

	return 0;
}

// Main program dialog callback
BOOL CALLBACK MainDlg(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	char buf[256];
	float freq;

	if (!hwndApp)
		hwndApp = hDlg;

	switch (iMsg)
	{
		case WM_INITDIALOG:
			// Set the main dialog's icon
			SetClassLong(hDlg, GCL_HICON, 
				(LONG) LoadImage(ghInstance, MAKEINTRESOURCE(IDI_ICON), IMAGE_ICON,
						 LR_DEFAULTSIZE, LR_DEFAULTSIZE, 0));
			// Add an always-on-top item to the system menu
			InsertMenu(GetSystemMenu(hDlg, 0), GetMenuItemCount(GetSystemMenu(hDlg, 0)) - 2, MF_STRING | MF_UNCHECKED | MF_BYPOSITION, SC_ON_TOP, "Always On-&Top");
			// Set the window size
			SetWindowPos(hDlg, NULL, 0, 0, 1030, 512, SWP_NOMOVE | SWP_SHOWWINDOW | SWP_NOZORDER);
			// Start the analysis thread
			_beginthread(analyze_thread, 0, NULL);
			return TRUE;

		case WM_CLOSE:
			// Signal termination of the analysis thread
			SetEvent(quitReqEvent);
			// Wait for it to finish.. up to one second
			WaitForSingleObject(quitAckEvent, 1000);
			// Close the dialog
			EndDialog(hDlg, 0);
			break;

		case WM_SYSCOMMAND:
		{
			switch (wParam)
			{
				case SC_ON_TOP:			// Handle always-on-top option
					if (AlwaysOnTop)
					{
						AlwaysOnTop = 0;
						CheckMenuItem(GetSystemMenu(hDlg, 0), SC_ON_TOP, MF_UNCHECKED);
						SetWindowPos(hDlg, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
					}
					else
					{
						AlwaysOnTop = 1;
						CheckMenuItem(GetSystemMenu(hDlg, 0), SC_ON_TOP, MF_CHECKED);
						SetWindowPos(hDlg, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
					}
					return FALSE;
			}
			break;
		}

		case WM_KEYDOWN:
			switch (wParam)
			{
				case 219: 		// Left bracket - decrease lock length by factor of 2
					if (RequestedLockLen > 64)
						RequestedLockLen /= 2;
					break;
				case 221:		// Right bracket - increase lock length by factor of 2
					RequestedLockLen *= 2; 
					break;
				case 111:		// Divide - decrease FFT size by factor of 2
					if (RequestedFFTSamples > 16)
						RequestedFFTSamples /= 2; 
					break;
				case 106:		// Multiply - increase FFT size by factor of 2
					RequestedFFTSamples *= 2; 
					break;
				case 109:		// Minus - decrease frequency multiplier by factor of 2
					freqMultiplier /= 2; 
					break;
				case 107:		// Plus - increase frequency multiplier by factor of 2
					freqMultiplier *= 2; 
					break;
				case 34:		// Page down - decrease filter samples by 1
					if (filterSamples > 0)
						filterSamples--;
					break;
				case 33:		// Page up - increase filter samples by 1
					filterSamples++;
					break;
				case 87:		// W - toggle drawing of raw waveforms
					drawWaves = !drawWaves;
					break;
				case 186:		// ; - toggle left channel log display
					logleft = !logleft;
					break;
				case 222:		// ' - toggle right channel log display
					logright = !logright;
					break;
				case 86:		// V - toggle drawing of voiceprint
					drawVoicePrint = !drawVoicePrint;
					break;
				case 65:		// A - increase voiceprint zoom
					voicePrintZoom *= 2.0f;
					break;
				case 90:		// Z - decrease voiceprint zoom
					if (voicePrintZoom > 1.0f)
						voicePrintZoom /= 2.0f;
					break;
				case 112:		// F1 - display help
					MessageBox(hDlg, helpText, "TSpec Help", MB_ICONINFORMATION);
					break;
				//default:
					//wsprintf(buf, "Unhandled keypress: %d", wParam);
					//MessageBox(hDlg, buf, "Moo!", 0);
			}
			break;

		case WM_LBUTTONDOWN:
		case WM_MOUSEMOVE:
			freq = LOWORD(lParam) * 44100.0f / RequestedFFTSamples;
			if (wParam & MK_LBUTTON)
			{
				// Left mouse button is being pushed or held down
				selectionMade = 0;
				if (!selectionInProgress)
				{
					selectionInProgress = 1;
					selectionStart = freq;
					sprintf(buf, "%s - %.0fHz", baseTitle, selectionStart);
				}
				else
					sprintf(buf, "%s - selecting %.0fHz to %.0fHz", baseTitle, selectionStart, selectionEnd);

				if (freq < selectionStart)
					selectionStart = freq;
				else
					selectionEnd = freq;
				// Display frequency in title bar when left mouse button is held down
				SetWindowText(hDlg, buf);
			}
			break;
		
		case WM_LBUTTONUP:
			freq = LOWORD(lParam) * 44100.0f / RequestedFFTSamples;

			// Left mouse button has been released
			selectionInProgress = 0;
			if (selectionStart != selectionEnd)
			{
				selectionMade = 1;
				sprintf(buf, "%s - selected %.0fHz to %.0fHz (mid %.0fHz)", baseTitle, selectionStart, selectionEnd, (selectionStart + selectionEnd) / 2);
				SetWindowText(hDlg, buf);
			}
			else
			{
				selectionMade = 0;
				// Restore original title bar when left mouse button is released
				SetWindowText(hDlg, baseTitle);
			}

			break;

		case WM_RBUTTONDOWN:
		case WM_RBUTTONDBLCLK:
			// Pause display when right mouse button is pressed
			pause_display = !pause_display;
			break;

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDCANCEL:
					EndDialog(hDlg, 0);
					return TRUE;
			}
	}

	return FALSE;
}

// Initialize DirectSoundCapture and all associated interfaces
int init_directsound_capture(void)
{
//	char buf[2048];

	// Create Win32 events
	quitReqEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	quitAckEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

    // Initialize COM
    if (FAILED(CoInitialize(NULL)))
        return 1;

	// Create the DirectSoundCapture object
	if (FAILED(DirectSoundCaptureCreate(NULL, &m_dsCap, NULL)))
		return 1;

	// Get device caps
	m_dsCapCaps.dwSize = sizeof(m_dsCapCaps);
	if (FAILED(m_dsCap->GetCaps(&m_dsCapCaps)))
		return 1;

	// Display device caps in a message box
	/*wsprintf(buf, 
		"Capture device information:\r\n\r\n"
		"Emulated: %s\r\n"
		"Certified: %s\r\n"
		"CD quality: %s\r\n"
		"Channels: %d",
		m_dsCapCaps.dwFlags & DSCCAPS_EMULDRIVER ? "Yes" : "No",
		m_dsCapCaps.dwFlags & DSCCAPS_CERTIFIED  ? "Yes" : "No",
		m_dsCapCaps.dwFormats & WAVE_FORMAT_4S16 ? "Yes" : "No",
		m_dsCapCaps.dwChannels);
	MessageBox(NULL, buf, "TSpec says... Caps!", MB_ICONINFORMATION);*/

	// Create capture buffer
	m_dsCapBufDesc.dwSize = sizeof(DSCBUFFERDESC);
	m_dsCapBufDesc.dwFlags = 0;
	m_dsCapBufDesc.dwBufferBytes = m_dsWaveFormat.nAvgBytesPerSec * 4;
	m_dsCapBufDesc.dwReserved = 0;
	m_dsCapBufDesc.lpwfxFormat = &m_dsWaveFormat;
	m_dsCapBufDesc.dwFXCount = 0;
	m_dsCapBufDesc.lpDSCFXDesc = NULL;
 	if (FAILED(m_dsCap->CreateCaptureBuffer(&m_dsCapBufDesc, &m_dsCapBuf, NULL)))
		return 1;

	// Get capture buffer caps
	m_dsCapBufCaps.dwSize = sizeof(DSCBCAPS);
	if (FAILED(m_dsCapBuf->GetCaps(&m_dsCapBufCaps)))
		return 1;

	// Display capture buffer caps in a message box
	/*wsprintf(buf,
		"Capture buffer information:\r\n\r\n"
		"Buffer length: %d",
		m_dsCapBufCaps.dwBufferBytes);
	MessageBox(NULL, buf, "TSpec says... Caps!", MB_ICONINFORMATION);*/
	
	// Save the capture buffer length
	capBufLen = m_dsCapBufCaps.dwBufferBytes;

	// Start capturing!
	m_dsCapBuf->Start(DSCBSTART_LOOPING);

	return 0;
}

// Shut down DirectSoundCapture and all associated interfaces, releasing any outstanding objects
void shutdown_directsound_capture(void)
{
	// Stop capturing if we still are
	if (m_dsCapBuf)
		m_dsCapBuf->Stop();

	// Release DirectX objects
	SAFE_RELEASE(m_dsCapBuf);
	SAFE_RELEASE(m_dsCap);

	// Uninitialize COM
    CoUninitialize();

	// Destroy Win32 events
	CloseHandle(quitReqEvent);
	CloseHandle(quitAckEvent);
}

signed short int *leftChannel, *rightChannel, *mixedBuffer;
DWORD FFTSamples = 0;		// Number of samples to run through the FFT

int capture_signal(void)
{
	unsigned int i;
	static DWORD lockLen = 0;				// Bytes locked
	DWORD readCursor;
	signed int lockOffset;
	VOID *firstBuf;
	VOID *secondBuf;
	DWORD firstLen, secondLen;
	DWORD lockSuccesses = 0, lockFailures = 0;

	// [Re]Allocate memory for DSound->sample buffers
	if (lockLen != RequestedLockLen)
	{
		lockLen = RequestedLockLen;
		lockSamples = lockLen / 2;
		channelSamples = lockLen / 4;
		leftChannel = (signed short int *) malloc(sizeof(signed short int) * lockLen);
		rightChannel = (signed short int *) malloc(sizeof(signed short int) * lockLen);
		mixedBuffer = (signed short int *) malloc(sizeof(signed short int) * lockLen);
	}

	// [Re]Allocate memory for sample->FFT->sample buffers
	if (FFTSamples != RequestedFFTSamples)
	{
		FFTSamples = RequestedFFTSamples;
		floatLeft = (float *) realloc(floatLeft, sizeof(float) * FFTSamples);
		floatRight = (float *) realloc(floatRight, sizeof(float) * FFTSamples);
		floatLeftOut = (float *) realloc(floatLeftOut, sizeof(float) * FFTSamples);
		floatRightOut = (float *) realloc(floatRightOut, sizeof(float) * FFTSamples);
		floatImag = (float *) realloc(floatImag, sizeof(float) * FFTSamples);
		floatImagOut = (float *) realloc(floatImagOut, sizeof(float) * FFTSamples);
		floatLeftMag = (float *) realloc(floatLeftMag, sizeof(float) * FFTSamples);
		floatRightMag = (float *) realloc(floatRightMag, sizeof(float) * FFTSamples);
		floatLeftMax = (float *) realloc(floatLeftMax, sizeof(float) * FFTSamples);
		floatRightMax = (float *) realloc(floatRightMax, sizeof(float) * FFTSamples);
		for (i = 0; i < FFTSamples; i++)
			floatLeftMax[i] = floatRightMax[i] = 1.0f;
	}

	// Get the position of the read cursor in the capture buffer
	// We can read everything up to this point (and not including it)
	m_dsCapBuf->GetCurrentPosition(NULL, &readCursor);

	// Calculate a starting point in the buffer behind the read cursor to lock at
	// Make sure it's aligned on a word boundary (just to be nice)
	readCursor -= readCursor % 4;
	if (readCursor < 0)						// Handle wrapping past the beginning of the buffer
		readCursor += capBufLen;
	lockOffset = readCursor - lockLen;// - 1;
	while (lockOffset < 0)						// Handle wrapping past the beginning of the buffer
		lockOffset += capBufLen;

	// Lock the capture buffer and grab some data, ah yeah!
	if (FAILED(m_dsCapBuf->Lock(lockOffset, lockLen, &firstBuf, &firstLen, &secondBuf, &secondLen, 0)))
	{
		//sprintf(buf, "Lock failed at: %d", lockOffset);
		//SetWindowText(hwndApp, buf);
		OutputDebugString("IDirectSoundCaptureBuffer8::Lock() failed!\n");
		lockFailures++;
		return 1;
	}

	//wsprintf(buf, "Locked %p for %d and %p for %d\n", firstBuf, firstLen, secondBuf, secondLen);
	//OutputDebugString(buf);
	lockSuccesses++;

	if (!pause_display)
	{
		// Copy the data out of the first buffer
		memcpy(mixedBuffer, firstBuf, firstLen);
		// Is there more (wrapped) data to copy?  If so, append it to the first chunk of copied data
		if (secondLen)
			memcpy(mixedBuffer + firstLen, secondBuf, secondLen);

		// Unlock the buffer
		if (FAILED(m_dsCapBuf->Unlock(firstBuf, firstLen, secondBuf, secondLen)))
		{
			OutputDebugString("IDirectSoundCaptureBuffer8::Unlock() failed!\n");
			return 1;
		}
	}

	return 0;
}

// Prepares the signal from DirectSoundCapture to be processed (by spectral analysis)
void prepare_signal(void)
{
	unsigned short int maxamp;
	float gain;
	unsigned int samplesToUse;
	unsigned int i, j;

	// Extract the left channel from the mixed buffer, finding a peak value
	j = 0;
	maxamp = mixedBuffer[0];
	for (i = 0; i < lockSamples; i += 2)
	{
		leftChannel[j] = mixedBuffer[i];
		if (abs(leftChannel[j]) > maxamp)
			maxamp = abs(leftChannel[j]);
		j++;
	}

	// Extract the right channel from the mixed buffer, finding a peak value
	j = 0;
	for (i = 1; i < lockSamples; i += 2)
	{
		rightChannel[j] = mixedBuffer[i];
		if (abs(rightChannel[j]) > maxamp)
			maxamp = abs(rightChannel[j]);
		j++;
	}

	// Determine a gain value based on the peak found
	if (maxamp)
		gain = 65536.0f / (float) maxamp;
	else
		gain = 1.0f;

	// Convert to floating point
	if (FFTSamples < channelSamples)
		samplesToUse = FFTSamples;
	else
		samplesToUse = channelSamples;
	for (i = 0; i < samplesToUse; i++)
	{
		floatLeft[i] = (float) leftChannel[i];
		floatRight[i] = (float) rightChannel[i];
		floatImag[i] = 0.0f;
	}
	// Fill in zeroes for everything after samples up to the FFT sample count
	for (i = channelSamples; i < FFTSamples; i++)
	{
		floatLeft[i] = 0.0f;
		floatRight[i] = 0.0f;
		floatImag[i] = 0.0f;
	}
}

float *floatLeftDisp;
float *floatRightDisp;

void process_output(void)
{
	float tmpf;
	static float log21 = 0.0f;
	unsigned int i, j;

	// Pre-calculate the log of 2.1f (used in log/decibel display)
	if (log21 == 0.0f)
		log21 = (float) log(2.1f);

	// *** LEFT CHANNEL ***
	// Convert to magnitude only - only necessary with FFT, not power spectrum
	//for (i = 0; i < FFTSamples; i++)
	//	floatLeftMag[i] = (float) (freqMultiplier * sqrt(floatLeftOut[i] * floatLeftOut[i] + floatImagOut[i] * floatImagOut[i]));

	// Multiply by frequency multiplier
	if (logleft)
	{
		// Use a logarithmic scale (decibels?)
		tmpf = freqMultiplier * 65536.0f;
		for (i = 0; i < FFTSamples; i++)
			floatLeftMag[i] = tmpf * (float) (log(0.0021f * sqrt(floatLeftMag[i]) + 2.1f) - log21);
	}
	else
	{
		// Use a linear scale (direct output from FFT / power spectrum)
		tmpf = freqMultiplier / 65536.0f;
		for (i = 0; i < FFTSamples; i++)
			floatLeftMag[i] *= tmpf;
	}

	// Filter the signal using a simple averaging filter going from n-f/2 to n+f/2
	/*if (filterSamples > 1)
	{
		ft = filterSamples / 2;
		for (i = ft; i < FFTSamples - ft; i++)
		{
			floatLeftOut[i] = 0.0f;
			for (j = i - ft; j < i + ft; j++)
				floatLeftOut[i] += floatLeftMag[j];
			floatLeftOut[i] /= filterSamples;
		}
		for (i = 0; i < ft; i++)
			floatLeftOut[i] = floatLeftMag[i];
		for (i = FFTSamples - ft; i < FFTSamples; i++)
			floatLeftOut[i] = floatLeftMag[i];
	}
	else
		for (i = 0; i < FFTSamples; i++)
			floatLeftOut[i] = floatLeftMag[i];*/

	// Filter the signal using a simple averaging filter going from n to n + f
	if (filterSamples)
	{
		for (i = 0; i < FFTSamples - filterSamples; i++)
		{
			floatLeftOut[i] = 0.0f;
			for (j = 0; j < filterSamples; j++)
				floatLeftOut[i] += floatLeftMag[i + j];
			floatLeftOut[i] /= filterSamples;
		}
		for (i = FFTSamples - filterSamples; i < FFTSamples; i++)
			floatLeftOut[i] = floatLeftMag[i];
		floatLeftDisp = floatLeftOut;
	}
	else
		floatLeftDisp = floatLeftMag;

	// *** RIGHT CHANNEL ***
	// Multiply by frequency multiplier
	if (logright)
	{
		// Use a logarithmic scale (decibels?)
		tmpf = freqMultiplier * 65536.0f;
		for (i = 0; i < FFTSamples; i++)
			floatRightMag[i] = tmpf * (float) (log(0.0021f * sqrt(floatRightMag[i]) + 2.1f) - log21);
	}
	else
	{
		// Use a linear scale (direct output from FFT / power spectrum)
		tmpf = freqMultiplier / 65536.0f;
		for (i = 0; i < FFTSamples; i++)
			floatRightMag[i] *= tmpf;
	}

	// Convert to magnitude only - only necessary with FFT, not power spectrum
	//for (i = 0; i < FFTSamples; i++)
	//	floatRightMag[i] = (float) (freqMultiplier * sqrt(floatRightOut[i] * floatRightOut[i] + floatImagOut[i] * floatImagOut[i]));

	// Filter the signal using a simple averaging filter going from n to n + f
	if (filterSamples)
	{
		for (i = 0; i < FFTSamples - filterSamples; i++)
		{
			floatRightOut[i] = 0.0f;
			for (j = 0; j < filterSamples; j++)
				floatRightOut[i] += floatRightMag[i + j];
			floatRightOut[i] /= filterSamples;
		}
		for (i = FFTSamples - filterSamples; i < FFTSamples; i++)
			floatRightOut[i] = floatRightMag[i];
		floatRightDisp = floatRightOut;
	}
	else
		floatRightDisp = floatRightMag;
}

DWORD frameTime, newFrameTime;
int frameCount = 0;
float fps = 0.0f;

void draw(void)
{
	char buf[1024];
	static RECT r, clientRect, gfxRect, plotRect;
	static HDC hdc = NULL;
	static HDC hdcBuffer = NULL;
	static HBITMAP hbmBuffer = NULL;
	static HGDIOBJ oldobj = NULL;
	static HPEN hpen = NULL;
	static unsigned int bufferWidth = 0, bufferHeight = 0;
	unsigned int w;
	signed int h, gfxRectWidth, gfxRectHeight;
	signed int x, y, baseline;
	int rcol, gcol, bcol;
	static HBRUSH hBackBrush = CreateSolidBrush(RGB(255, 255, 255));
	int selectionStartX, selectionEndX;
	unsigned int i;

	// Draw!!  (use a double-buffered drawing system)
	GetClientRect(hwndApp, &clientRect);
	w = clientRect.right - clientRect.left;
	h = (clientRect.bottom - clientRect.top);
	if (w != bufferWidth || (unsigned int) h != bufferHeight)
	{
		// Recreate the double buffer, we've resized
		if (!hdcBuffer)
			hdcBuffer = CreateCompatibleDC(NULL);
		bufferWidth = w;
		bufferHeight = h;
		if (hbmBuffer)
			DeleteObject(hbmBuffer);
		hdc = GetDC(hwndApp);
		hbmBuffer = CreateCompatibleBitmap(hdc, bufferWidth, bufferHeight);
		ReleaseDC(hwndApp, hdc);
		SelectObject(hdcBuffer, hbmBuffer);
	}

	// Calculate rects / dimensions for layout
	gfxRect.top = clientRect.top + 20;
	gfxRect.left = clientRect.left;
	gfxRect.right = clientRect.right;
	gfxRect.bottom = clientRect.bottom;
	gfxRectWidth = gfxRect.right - gfxRect.left;
	gfxRectHeight = gfxRect.bottom - gfxRect.top;

	//h /= 2;
	//baseline = h / 2;

	hdc = hdcBuffer;//GetDC(hwndApp);
	if (!hdc)
		return;

	if (drawVoicePrint)
	{
		// Clear the text area's background to white
		SelectObject(hdc, hBackBrush);
		Rectangle(hdc, clientRect.left, clientRect.top, clientRect.right, gfxRect.top);

		// Scroll all of the old stuff left by one pixel
		x = gfxRect.right - 1;
		BitBlt(hdc, gfxRect.left, gfxRect.top, gfxRectWidth - 1, gfxRectHeight, hdcBuffer, gfxRect.left + 1, gfxRect.top, SRCCOPY);
		
		// Calculations for how we'll iterate through the frequency data
		float freqIndex = 0.0f;
		float freqIndexInc = (((float) FFTSamples) / voicePrintZoom / 2.0f) / gfxRectHeight;
		float maxFreqStrength = 65536.0f * 16.0f;
		float freqStrength;

		// Loop from top to bottom, iterating through the frequency data...
		for (y = gfxRect.top; y < gfxRect.bottom; y++)
		{
			// Calculate the overall signal strength for this particular frequency (combination of left + right)
			freqStrength = ((floatLeftMag[(int) freqIndex] + floatRightMag[(int) freqIndex]) / maxFreqStrength);
			freqIndex += freqIndexInc;

			// Calculate a color for the pixel from the signal strength (simple three-color linear gradient)
			if (freqStrength < 0.5f)
			{
				rcol = (int) (freqStrength * 2 * 255.0f);
				gcol = 0;
				bcol = 0;
			}
			else if (freqStrength < 1.0f)
			{
				rcol = 255;
				gcol = (int) ((freqStrength - 0.5f) * 2 * 255.0f);
				bcol = 0;
			}
			else if (freqStrength < 4.0f)
			{
				rcol = 255;
				gcol = 255;
				bcol = (int) ((freqStrength - 1.0f) / 3.0f * 255.0f);
			}
			else
			{
				rcol = 255;
				gcol = 255;
				bcol = 255;
			}

			// Draw the pixel!
			SetPixel(hdc, x, gfxRect.top * 2 + gfxRectHeight - y, RGB(rcol, gcol, bcol));
		}
	}
	else
	{
		// Clear the whole background to white
		SelectObject(hdc, hBackBrush);
		Rectangle(hdc, clientRect.left, clientRect.top, clientRect.right, clientRect.bottom);

		// Draw borders
		//MoveToEx(hdc, 0, h, NULL);
		//LineTo(hdc, w, h);
		//MoveToEx(hdc, w / 2, 0, NULL);
		//LineTo(hdc, w / 2, h * 2);

		// LEFT CHANNEL

		// Draw spectrum analysis (raw magnitude, unfiltered)
		if (pause_display && filterSamples > 0)
		{
			hpen = CreatePen(PS_SOLID, 0, RGB(255, 192, 192));
			oldobj = SelectObject(hdc, hpen);
			plotRect.top = gfxRect.top;
			plotRect.left = gfxRect.left;
			plotRect.right = plotRect.left + (gfxRect.right - gfxRect.left) / 2;
			if (!drawWaves)
				plotRect.right = gfxRect.right;
			plotRect.bottom = plotRect.top + (gfxRect.bottom - gfxRect.top) / 2;
			h = plotRect.bottom - plotRect.top;
			baseline = h / 2;
			y = plotRect.top + h - ((int) floatLeftMag[0] * baseline) / FREQ_DIVISOR;
			if (y < plotRect.top)
				y = plotRect.top;
			MoveToEx(hdc, plotRect.left, y, NULL);
			w = plotRect.right - plotRect.left;
			if ((FFTSamples / 2) < w)
				w = FFTSamples / 2;
			for (i = 1; i < w; i++)
			{
				y = plotRect.top + h - ((int) floatLeftMag[i] * baseline) / FREQ_DIVISOR;
				if (y < plotRect.top)
					y = plotRect.top;
				if (y > plotRect.bottom)
					y = plotRect.bottom;
				LineTo(hdc, plotRect.left + i, y);
			}
			SelectObject(hdc, oldobj);
			DeleteObject(hpen);
		}

		// Draw spectrum analysis (filtered)
		if (pause_display && filterSamples > 0)
		{
			hpen = CreatePen(PS_SOLID, 2, RGB(127, 0, 0));
			oldobj = SelectObject(hdc, hpen);
		}
		plotRect.top = gfxRect.top;
		plotRect.left = gfxRect.left;
		plotRect.right = plotRect.left + (gfxRect.right - gfxRect.left) / 2;
		if (!drawWaves)
			plotRect.right = gfxRect.right;
		plotRect.bottom = plotRect.top + (gfxRect.bottom - gfxRect.top) / 2;
		h = plotRect.bottom - plotRect.top;
		baseline = h / 2;
		y = plotRect.top + h - ((int) floatLeftDisp[0] * baseline) / FREQ_DIVISOR;
		if (y < plotRect.top)
			y = plotRect.top;
		MoveToEx(hdc, plotRect.left, y, NULL);
		w = plotRect.right - plotRect.left;
		if ((FFTSamples / 2) < w)
			w = FFTSamples / 2;
		for (i = 1; i < w; i++)
		{
			y = plotRect.top + h - ((int) floatLeftDisp[i] * baseline) / FREQ_DIVISOR;
			if (y < plotRect.top)
				y = plotRect.top;
			if (y > plotRect.bottom)
				y = plotRect.bottom;
			LineTo(hdc, plotRect.left + i, y);
		}
		if (hpen)
		{
			SelectObject(hdc, oldobj);
			DeleteObject(hpen);
			hpen = NULL;
		}

		// Draw waveform
		if (drawWaves)
		{
			plotRect.top = gfxRect.top;
			plotRect.left = gfxRect.left + (gfxRect.right - gfxRect.left) / 2;
			plotRect.right = gfxRect.right;
			plotRect.bottom = plotRect.top + (gfxRect.bottom - gfxRect.top) / 2;
			h = plotRect.bottom - plotRect.top;
			baseline = plotRect.top + h / 2;
			MoveToEx(hdc, plotRect.left, baseline + (leftChannel[0] * h) / (65536), NULL);
			w = plotRect.right - plotRect.left;
			if (channelSamples < w)
				w = channelSamples;
			for (i = 1; i < w; i++)
				LineTo(hdc, plotRect.left + i, baseline + (leftChannel[i] * h) / (65536));
		}

		// RIGHT CHANNEL

		// Draw spectrum analysis (raw magnitude, unfiltered)
		if (pause_display && filterSamples > 0)
		{
			hpen = CreatePen(PS_SOLID, 0, RGB(255, 192, 192));
			oldobj = SelectObject(hdc, hpen);
			plotRect.top = gfxRect.top + (gfxRect.bottom - gfxRect.top) / 2;
			plotRect.left = gfxRect.left;
			plotRect.right = plotRect.left + (gfxRect.right - gfxRect.left) / 2;
			if (!drawWaves)
				plotRect.right = gfxRect.right;
			plotRect.bottom = plotRect.top + (gfxRect.bottom - gfxRect.top) / 2;
			h = plotRect.bottom - plotRect.top;
			baseline = h / 2;
			y = plotRect.top + h - ((int) floatRightMag[0] * baseline) / FREQ_DIVISOR;
			if (y < plotRect.top)
				y = plotRect.top;
			MoveToEx(hdc, plotRect.left, y, NULL);
			w = plotRect.right - plotRect.left;
			if ((FFTSamples / 2) < w)
				w = FFTSamples / 2;
			for (i = 1; i < w; i++)
			{
				y = plotRect.top + h - ((int) floatRightMag[i] * baseline) / FREQ_DIVISOR;
				if (y < plotRect.top)
					y = plotRect.top;
				LineTo(hdc, plotRect.left + i, y);
			}
			SelectObject(hdc, oldobj);
			DeleteObject(hpen);
		}

		// Draw spectrum analysis (filtered)
		if (pause_display && filterSamples > 0)
		{
			hpen = CreatePen(PS_SOLID, 2, RGB(127, 0, 0));
			oldobj = SelectObject(hdc, hpen);
		}
		plotRect.top = gfxRect.top + (gfxRect.bottom - gfxRect.top) / 2;
		plotRect.left = gfxRect.left;
		plotRect.right = plotRect.left + (gfxRect.right - gfxRect.left) / 2;
		if (!drawWaves)
			plotRect.right = gfxRect.right;
		plotRect.bottom = plotRect.top + (gfxRect.bottom - gfxRect.top) / 2;
		h = plotRect.bottom - plotRect.top;
		baseline = h / 2;
		y = plotRect.top + h - ((int) floatRightDisp[0] * baseline) / FREQ_DIVISOR;
		if (y < plotRect.top)
			y = plotRect.top;
		MoveToEx(hdc, plotRect.left, y, NULL);
		w = plotRect.right - plotRect.left;
		if ((FFTSamples / 2) < w)
			w = FFTSamples / 2;
		for (i = 1; i < w; i++)
		{
			y = plotRect.top + h - ((int) floatRightDisp[i] * baseline) / FREQ_DIVISOR;
			if (y < plotRect.top)
				y = plotRect.top;
			LineTo(hdc, plotRect.left + i, y);
		}
		if (hpen)
		{
			SelectObject(hdc, oldobj);
			DeleteObject(hpen);
			hpen = NULL;
		}
		
		// Draw waveform
		if (drawWaves)
		{
			plotRect.top = gfxRect.top + (gfxRect.bottom - gfxRect.top) / 2;
			plotRect.left = gfxRect.left + (gfxRect.right - gfxRect.left) / 2;
			plotRect.right = gfxRect.right;
			plotRect.bottom = plotRect.top + (gfxRect.bottom - gfxRect.top) / 2;
			h = plotRect.bottom - plotRect.top;
			baseline = plotRect.top + h / 2;
			MoveToEx(hdc, plotRect.left, baseline + (rightChannel[0] * h) / (65536), NULL);
			w = plotRect.right - plotRect.left;
			if (channelSamples < w)
				w = channelSamples;
			for (i = 1; i < w; i++)
				LineTo(hdc, plotRect.left + i, baseline + (rightChannel[i] * h) / (65536));
		}

		// Show selection region
		if (selectionMade || selectionInProgress)
		{
			selectionStartX = (int) ((selectionStart * RequestedFFTSamples) / 44100.0f);
			selectionEndX = (int) ((selectionEnd * RequestedFFTSamples) / 44100.0f);
			if (selectionStartX != selectionEndX)
			{
				h = gfxRect.bottom - gfxRect.top;
				r.left = selectionStartX;
				r.right = selectionEndX;
				r.top = gfxRect.top;
				r.bottom = gfxRect.bottom - 1;
				InvertRect(hdc, &r);
			}
		}
	}

	// Display the frame rate in the upper right corner
	//sprintf(buf, "(%dms) - %.2f fps", lastFrameTime, fps);
	sprintf(buf, "%.2f fps", fps);
	r.left = clientRect.right - 200;
	r.top = 1;
	r.bottom = 30;
	r.right = clientRect.right - 2;
	DrawText(hdc, buf, strlen(buf), &r, DT_RIGHT);
	
	// Display operational parameters in the upper left corner
	sprintf(buf, "Sample len: %d   FFT len: %d   Filter len: %d   Freq mult: %.3f   VP zoom: %.2f   (F1 for Help)", 
		channelSamples, FFTSamples, filterSamples, freqMultiplier, voicePrintZoom);
	r.left = clientRect.left + 2;
	r.top = 1;
	r.bottom = 30;
	r.right = clientRect.right - 1;
	DrawText(hdc, buf, strlen(buf), &r, DT_LEFT);

	// Blit the double buffer to the screen
	hdc = GetDC(hwndApp);
	BitBlt(hdc, 0, 0, bufferWidth, bufferHeight, hdcBuffer, 0, 0, SRCCOPY);
	ReleaseDC(hwndApp, hdc);
}

// Thread to grab data from the buffer, analyze it, and update any displays
void _cdecl analyze_thread(void *ptr)
{
	DWORD oneFrameStart, oneFrameEnd;

	// Set the thread priority to low
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_LOWEST);

	// Ask for 1ms timer resolution and take an initial timer reading
	timeBeginPeriod(1);
	frameTime = timeGetTime();

	// Infinite analysis loop
	while (1)
	{
		// Wait a little while and break if the quit request is signalled
		// This acts as a primitive framerate limiter...
		//if (WaitForSingleObject(quitReqEvent, 5) == WAIT_OBJECT_0)
		if (WaitForSingleObject(quitReqEvent, 0) == WAIT_OBJECT_0)
			break;

		oneFrameStart = timeGetTime();

		// Don't proceed if the display is paused
		if (pause_display)
		{
			Sleep(5);
			continue;
		}

		// Capture the signal using DirectSoundCapture
		if (capture_signal())
			continue;

		// Prepare the signal for spectral analysis - it needs to be:
		// 1) Separated into individual channels (left, right)
		// 2) In floating point format
		prepare_signal();

		// Run FFT!!!!!!!!!!!!!!!!!!!!!! (LEFT CHANNEL)
		//fft_float(FFTSamples, FALSE, floatLeft, floatImag, floatLeftOut, floatImagOut);

		// Use power spectrum instead, for increased performance
		PowerSpectrum(FFTSamples, floatLeft, floatLeftMag);

		// Run FFT!!!!!!!!!!!!!!!!!!!!!! (RIGHT CHANNEL)
		//fft_float(FFTSamples, FALSE, floatRight, floatImag, floatRightOut, floatImagOut);

		// Use power spectrum instead, for increased performance
		PowerSpectrum(FFTSamples, floatRight, floatRightMag);

		// Process the output of the FFT, in preparation for visualization
		process_output();

		// Increment the frame counter
		frameCount++;
		newFrameTime = timeGetTime();
		// Calculate the framerate every now and then
		if (newFrameTime - frameTime >= 100)
		{
			fps = (float) frameCount / (newFrameTime - frameTime) * 1000;
			frameTime = newFrameTime;
			frameCount = 0;
		}

		// Draw the results
		draw();

		oneFrameEnd = timeGetTime();
		lastFrameTime = oneFrameEnd - oneFrameStart;

		// Limit the framerate
		unsigned int targetFrameRate = 150;
		if (lastFrameTime < (1000 / targetFrameRate))
			Sleep((1000 / targetFrameRate) - lastFrameTime);
	}

	// We're done with our 1ms timer resolution
	timeEndPeriod(1);

	// Signal completion of the analysis thread
	SetEvent(quitAckEvent);
}
